@@ -1,3 +1,4 @@
+import logging
 import os
 import time
 import threading
@@ -9,23 +10,31 @@ import functools
 import itertools
 import hashlib
 import weakref
+import queue
 from collections import defaultdict
-try:
-    # Python 3.8+ https://docs.python.org/3/whatsnew/3.7.html#id3
-    from collections.abc import MutableMapping
-except ImportError:
-    # Python 2.6, 2.7
-    from collections import MutableMapping
+from collections.abc import MutableMapping
 
 import six
-from six.moves import queue
 import redis
 
 from ._zset import ZSet
 
 
+LOGGER = logging.getLogger('fakeredis')
+REDIS_LOG_LEVELS = {
+    b'LOG_DEBUG': 0,
+    b'LOG_VERBOSE': 1,
+    b'LOG_NOTICE': 2,
+    b'LOG_WARNING': 3
+}
+REDIS_LOG_LEVELS_TO_LOGGING = {
+    0: logging.DEBUG,
+    1: logging.INFO,
+    2: logging.INFO,
+    3: logging.WARNING
+}
+
 MAX_STRING_SIZE = 512 * 1024 * 1024
-INF = float('inf')
 
 INVALID_EXPIRE_MSG = "invalid expire time in {}"
 WRONGTYPE_MSG = \
@@ -65,17 +74,20 @@ BAD_SUBCOMMAND_MSG = "Unknown {} subcommand or wrong # of args."
 BAD_COMMAND_IN_PUBSUB_MSG = \
     "only (P)SUBSCRIBE / (P)UNSUBSCRIBE / PING / QUIT allowed in this context"
 CONNECTION_ERROR_MSG = "FakeRedis is emulating a connection error."
+REQUIRES_MORE_ARGS = "{} requires {} arguments or more."
+LOG_WRONG_FIRST_ARG = "First argument must be a number (log level)."
+LOG_INVALID_DEBUG_LEVEL = "Invalid debug level."
 
 FLAG_NO_SCRIPT = 's'      # Command not allowed in scripts
 
 
-class SimpleString(object):
+class SimpleString:
     def __init__(self, value):
         assert isinstance(value, bytes)
         self.value = value
 
 
-class NoResponse(object):
+class NoResponse:
     """Returned by pub/sub commands to indicate that no response should be returned"""
     pass
 
@@ -86,25 +98,6 @@ PONG = SimpleString(b'PONG')
 BGSAVE_STARTED = SimpleString(b'Background saving started')
 
 
-if six.PY2:
-    def isfinite(value):
-        return not math.isinf(value) and not math.isnan(value)
-
-    # Not the same as six.byte2int, which takes a bytes in both Python 2+3.
-    # This takes an integer in Python 3.
-    def byte_to_int(value):
-        return ord(value)
-else:
-    def isfinite(value):
-        return math.isfinite(value)
-
-    def byte_to_int(value):
-        assert isinstance(value, int)
-        return value
-
-    long = int
-
-
 def null_terminate(s):
     # Redis uses C functions on some strings, which means they stop at the
     # first NULL.
@@ -188,7 +181,7 @@ def compile_pattern(pattern):
     return re.compile(regex, re.S)
 
 
-class Item(object):
+class Item:
     """An item stored in the database"""
 
     __slots__ = ['value', 'expireat']
@@ -198,7 +191,7 @@ class Item(object):
         self.expireat = None
 
 
-class CommandItem(object):
+class CommandItem:
     """An item referenced by a command.
 
     It wraps an Item but has extra fields to manage updates and notifications.
@@ -334,7 +327,7 @@ class Hash(dict):
     redis_type = b'hash'
 
 
-class Int(object):
+class Int:
     """Argument converter for 64-bit signed integers"""
 
     DECODE_ERROR = INVALID_INT_MSG
@@ -350,7 +343,7 @@ class Int(object):
     def decode(cls, value):
         try:
             out = int(value)
-            if not cls.valid(out) or six.ensure_binary(str(out)) != value:
+            if not cls.valid(out) or str(out).encode() != value:
                 raise ValueError
         except ValueError:
             raise redis.ResponseError(cls.DECODE_ERROR)
@@ -359,7 +352,7 @@ class Int(object):
     @classmethod
     def encode(cls, value):
         if cls.valid(value):
-            return six.ensure_binary(str(value))
+            return str(value).encode()
         else:
             raise redis.ResponseError(cls.ENCODE_ERROR)
 
@@ -393,7 +386,7 @@ class Timeout(Int):
     MIN_VALUE = 0
 
 
-class Float(object):
+class Float:
     """Argument converter for floating-point values.
 
     Redis uses long double for some cases (INCRBYFLOAT, HINCRBYFLOAT)
@@ -427,7 +420,7 @@ class Float(object):
                 # Values that over- or underflow- are explicitly rejected by
                 # redis. This is a crude hack to determine whether the input
                 # may have been such a value.
-                if out in (INF, -INF, 0.0) and re.match(b'^[^a-zA-Z]*[1-9]', value):
+                if out in (math.inf, -math.inf, 0.0) and re.match(b'^[^a-zA-Z]*[1-9]', value):
                     raise ValueError
             return out
         except ValueError:
@@ -436,14 +429,14 @@ class Float(object):
     @classmethod
     def encode(cls, value, humanfriendly):
         if math.isinf(value):
-            return six.ensure_binary(str(value))
+            return str(value).encode()
         elif humanfriendly:
             # Algorithm from ld2string in redis
             out = '{:.17f}'.format(value)
             out = re.sub(r'(?:\.)?0+$', '', out)
-            return six.ensure_binary(out)
+            return out.encode()
         else:
-            return six.ensure_binary('{:.17g}'.format(value))
+            return '{:.17g}'.format(value).encode()
 
 
 class SortFloat(Float):
@@ -451,11 +444,11 @@ class SortFloat(Float):
 
     @classmethod
     def decode(cls, value):
-        return super(SortFloat, cls).decode(
+        return super().decode(
             value, allow_leading_whitespace=True, allow_empty=True, crop_null=True)
 
 
-class ScoreTest(object):
+class ScoreTest:
     """Argument converter for sorted set score endpoints."""
     def __init__(self, value, exclusive=False):
         self.value = value
@@ -490,7 +483,7 @@ class ScoreTest(object):
         return (self.value, BeforeAny() if self.exclusive else AfterAny())
 
 
-class StringTest(object):
+class StringTest:
     """Argument converter for sorted set LEX endpoints."""
     def __init__(self, value, exclusive):
         self.value = value
@@ -511,7 +504,7 @@ class StringTest(object):
 
 
 @functools.total_ordering
-class BeforeAny(object):
+class BeforeAny:
     def __gt__(self, other):
         return False
 
@@ -520,7 +513,7 @@ class BeforeAny(object):
 
 
 @functools.total_ordering
-class AfterAny(object):
+class AfterAny:
     def __lt__(self, other):
         return False
 
@@ -528,7 +521,7 @@ class AfterAny(object):
         return isinstance(other, AfterAny)
 
 
-class Key(object):
+class Key:
     """Marker to indicate that argument in signature is a key"""
     UNSPECIFIED = object()
 
@@ -537,7 +530,7 @@ class Key(object):
         self.missing_return = missing_return
 
 
-class Signature(object):
+class Signature:
     def __init__(self, name, fixed, repeat=(), flags=""):
         self.name = name
         self.fixed = fixed
@@ -596,7 +589,7 @@ def valid_response_type(value, nested=False):
     if isinstance(value, NoResponse) and not nested:
         return True
     if value is not None and not isinstance(value, (bytes, SimpleString, redis.ResponseError,
-                                                    int, long, list)):
+                                                    int, list)):
         return False
     if isinstance(value, list):
         if any(not valid_response_type(item, True) for item in value):
@@ -613,7 +606,7 @@ def command(*args, **kwargs):
     return decorator
 
 
-class FakeServer(object):
+class FakeServer:
     def __init__(self):
         self.lock = threading.Lock()
         self.dbs = defaultdict(lambda: Database(self.lock))
@@ -626,7 +619,7 @@ class FakeServer(object):
         self.connected = True
 
 
-class FakeSocket(object):
+class FakeSocket:
     def __init__(self, server):
         self._server = server
         self._db = server.dbs[0]
@@ -726,8 +719,6 @@ class FakeSocket(object):
             return [self._decode_result(r) for r in result]
         elif isinstance(result, SimpleString):
             return result.value
-        elif six.PY2 and isinstance(result, int):
-            return long(result)
         else:
             return result
 
@@ -1016,9 +1007,8 @@ class FakeSocket(object):
             raise redis.ResponseError(SRC_DST_SAME_MSG)
         if not key or key.key in self._server.dbs[db]:
             return 0
-        # TODO: what is the interaction with expiry and WATCH?
-        self._server.dbs[db][key.key] = self._server.dbs[self._db_num][key.key]
-        key.value = None   # Causes deletion
+        # TODO: what is the interaction with expiry?
+        self._server.dbs[db][key.key] = self._server.dbs[self._db_num].pop(key.key)
         return 1
 
     @command(())
@@ -1268,7 +1258,7 @@ class FakeSocket(object):
             value = key.value[start:end]
         else:
             value = key.value
-        return sum([bin(byte_to_int(l)).count('1') for l in value])
+        return bin(int.from_bytes(value, 'little')).count('1')
 
     @command((Key(bytes), Int))
     def decrby(self, key, amount):
@@ -1292,7 +1282,7 @@ class FakeSocket(object):
     def incrbyfloat(self, key, amount):
         # TODO: introduce convert_order so that we can specify amount is Float
         c = Float.decode(key.get(b'0')) + Float.decode(amount)
-        if not isfinite(c):
+        if not math.isfinite(c):
             raise redis.ResponseError(NONFINITE_MSG)
         encoded = Float.encode(c, True)
         key.update(encoded)
@@ -1309,7 +1299,7 @@ class FakeSocket(object):
         remaining = offset % 8
         actual_bitoffset = 7 - remaining
         try:
-            actual_val = byte_to_int(value[byte])
+            actual_val = value[byte]
         except IndexError:
             return 0
         return 1 if (1 << actual_bitoffset) & actual_val else 0
@@ -1325,7 +1315,7 @@ class FakeSocket(object):
             # bit.
             needed = byte - (len(val) - 1)
             val += b'\x00' * needed
-        old_byte = byte_to_int(val[byte])
+        old_byte = val[byte]
         if value == 1:
             new_byte = old_byte | (1 << actual_bitoffset)
         else:
@@ -1490,7 +1480,7 @@ class FakeSocket(object):
     @command((Key(Hash), bytes, bytes))
     def hincrbyfloat(self, key, field, amount):
         c = Float.decode(key.value.get(field, b'0')) + Float.decode(amount)
-        if not isfinite(c):
+        if not math.isfinite(c):
             raise redis.ResponseError(NONFINITE_MSG)
         encoded = Float.encode(c, True)
         key.value[field] = encoded
@@ -1746,15 +1736,27 @@ class FakeSocket(object):
     def scard(self, key):
         return len(key.value)
 
-    def _setop(self, op, stop_when_empty, dst, key, *keys):
+    def _calc_setop(self, op, stop_if_missing, key, *keys):
+        if stop_if_missing and not key.value:
+            return set()
         ans = key.value.copy()
         for other in keys:
-            if stop_when_empty and not ans:
-                break
             value = other.value if other.value is not None else set()
             if not isinstance(value, set):
                 raise redis.ResponseError(WRONGTYPE_MSG)
+            if stop_if_missing and not value:
+                return set()
             ans = op(ans, value)
+        return ans
+
+    def _setop(self, op, stop_if_missing, dst, key, *keys):
+        """Apply one of SINTER[STORE], SUNION[STORE], SDIFF[STORE].
+
+        If `stop_if_missing`, the output will be made an empty set as soon as
+        an empty input set is encountered (use for SINTER[STORE]). May assume
+        that `key` is a set (or empty), but `keys` could be anything.
+        """
+        ans = self._calc_setop(op, stop_if_missing, key, *keys)
         if dst is None:
             return list(ans)
         else:
@@ -1769,7 +1771,9 @@ class FakeSocket(object):
     def sdiffstore(self, dst, *keys):
         return self._setop(lambda a, b: a - b, False, dst, *keys)
 
-    @command((Key(set, missing_return=[]),), (Key(),))
+    # The following keys can't be marked as sets because of the
+    # stop_if_empty early-out.
+    @command((Key(set),), (Key(),))
     def sinter(self, *keys):
         return self._setop(lambda a, b: a & b, True, None, *keys)
 
@@ -2157,11 +2161,19 @@ class FakeSocket(object):
             else:
                 out_members.intersection_update(s)
 
-        out = ZSet()
-        for s, w in zip(sets, weights):
+        # We first build a regular dict and turn it into a ZSet. The
+        # reason is subtle: a ZSet won't update a score from -0 to +0
+        # (or vice versa) through assignment, but a regular dict will.
+        out = {}
+        # The sort affects the order of floating-point operations.
+        # Note that redis uses qsort(1), which has no stability guarantees,
+        # so we can't be sure to match it in all cases.
+        for s, w in sorted(zip(sets, weights), key=lambda x: len(x[0])):
             for member, score in s.items():
                 score *= w
-                if math.isnan(score):
+                # Redis only does this step for ZUNIONSTORE. See
+                # https://github.com/antirez/redis/issues/3954.
+                if func == 'ZUNIONSTORE' and math.isnan(score):
                     score = 0.0
                 if member not in out_members:
                     continue
@@ -2177,10 +2189,16 @@ class FakeSocket(object):
                         score = min(old, score)
                     else:
                         assert False     # pragma: nocover
+                if math.isnan(score):
+                    score = 0.0
                 out[member] = score
 
-        dest.value = out
-        return len(out)
+        out_zset = ZSet()
+        for member, score in out.items():
+            out_zset[member] = score
+
+        dest.value = out_zset
+        return len(out_zset)
 
     @command((Key(), Int, bytes), (bytes,))
     def zunionstore(self, dest, numkeys, *args):
@@ -2235,15 +2253,15 @@ class FakeSocket(object):
 
     def _convert_redis_arg(self, lua_runtime, value):
         if isinstance(value, bytes):
-            return six.ensure_binary(value)
-        elif isinstance(value, (int, long, float)):
-            return six.ensure_binary('{:.17g}'.format(value))
+            return value
+        elif isinstance(value, (int, float)):
+            return '{:.17g}'.format(value).encode()
         else:
             # TODO: add a constant for this, and add the context
             raise redis.ResponseError('Lua redis() command arguments must be strings or integers')
 
     def _convert_redis_result(self, lua_runtime, result):
-        if isinstance(result, (bytes, int, long)):
+        if isinstance(result, (bytes, int)):
             return result
         elif isinstance(result, SimpleString):
             return lua_runtime.table_from({b"ok": result.value})
@@ -2283,8 +2301,8 @@ class FakeSocket(object):
                 item = result[index]
                 result_list.append(self._convert_lua_result(item))
             return result_list
-        elif isinstance(result, six.text_type):
-            return six.ensure_binary(result)
+        elif isinstance(result, str):
+            return result.encode()
         elif isinstance(result, float):
             return int(result)
         elif isinstance(result, bool):
@@ -2312,24 +2330,39 @@ class FakeSocket(object):
         except Exception as ex:
             return lua_runtime.table_from({b"err": str(ex)})
 
+    def _lua_redis_log(self, lua_runtime, expected_globals, lvl, *args):
+        self._check_for_lua_globals(lua_runtime, expected_globals)
+        if len(args) < 1:
+            raise redis.ResponseError(REQUIRES_MORE_ARGS.format("redis.log()", "two"))
+        if lvl not in REDIS_LOG_LEVELS.values():
+            raise redis.ResponseError(LOG_INVALID_DEBUG_LEVEL)
+        msg = ' '.join([x.decode('utf-8')
+                        if isinstance(x, bytes) else str(x)
+                        for x in args if not isinstance(x, bool)])
+        LOGGER.log(REDIS_LOG_LEVELS_TO_LOGGING[lvl], msg)
+
     @command((bytes, Int), (bytes,), flags='s')
     def eval(self, script, numkeys, *keys_and_args):
-        from lupa import LuaRuntime, LuaError
+        from lupa import LuaRuntime, LuaError, as_attrgetter
 
         if numkeys > len(keys_and_args):
             raise redis.ResponseError(TOO_MANY_KEYS_MSG)
         if numkeys < 0:
             raise redis.ResponseError(NEGATIVE_KEYS_MSG)
-        sha1 = six.ensure_binary(hashlib.sha1(script).hexdigest())
+        sha1 = hashlib.sha1(script).hexdigest().encode()
         self._server.script_cache[sha1] = script
         lua_runtime = LuaRuntime(encoding=None, unpack_returned_tuples=True)
 
         set_globals = lua_runtime.eval(
             """
-            function(keys, argv, redis_call, redis_pcall)
+            function(keys, argv, redis_call, redis_pcall, redis_log, redis_log_levels)
                 redis = {}
                 redis.call = redis_call
                 redis.pcall = redis_pcall
+                redis.log = redis_log
+                for level, pylevel in python.iterex(redis_log_levels.items()) do
+                    redis[level] = pylevel
+                end
                 redis.error_reply = function(msg) return {err=msg} end
                 redis.status_reply = function(msg) return {ok=msg} end
                 KEYS = keys
@@ -2342,7 +2375,9 @@ class FakeSocket(object):
             lua_runtime.table_from(keys_and_args[:numkeys]),
             lua_runtime.table_from(keys_and_args[numkeys:]),
             functools.partial(self._lua_redis_call, lua_runtime, expected_globals),
-            functools.partial(self._lua_redis_pcall, lua_runtime, expected_globals)
+            functools.partial(self._lua_redis_pcall, lua_runtime, expected_globals),
+            functools.partial(self._lua_redis_log, lua_runtime, expected_globals),
+            as_attrgetter(REDIS_LOG_LEVELS)
         )
         expected_globals.update(lua_runtime.globals().keys())
 
@@ -2369,7 +2404,7 @@ class FakeSocket(object):
             if len(args) != 1:
                 raise redis.ResponseError(BAD_SUBCOMMAND_MSG.format('SCRIPT'))
             script = args[0]
-            sha1 = six.ensure_binary(hashlib.sha1(script).hexdigest())
+            sha1 = hashlib.sha1(script).hexdigest().encode()
             self._server.script_cache[sha1] = script
             return sha1
         else:
@@ -2447,7 +2482,7 @@ setattr(FakeSocket, 'exec', FakeSocket.exec_)
 delattr(FakeSocket, 'exec_')
 
 
-class _DummyParser(object):
+class _DummyParser:
     def __init__(self, socket_read_size):
         self.socket_read_size = socket_read_size
 
@@ -2462,7 +2497,7 @@ class _DummyParser(object):
 try:
     from redis.selector import BaseSelector
 except ImportError:
-    class BaseSelector(object):
+    class BaseSelector:
         def __init__(self, sock):
             self.sock = sock
 
@@ -2492,15 +2527,18 @@ class FakeSelector(BaseSelector):
 class FakeConnection(redis.Connection):
     description_format = "FakeConnection<db=%(db)s>"
 
-    def __init__(self, server, db=0, password=None,
+    def __init__(self, server, db=0, username=None, password=None,
                  socket_timeout=None, socket_connect_timeout=None,
                  socket_keepalive=False, socket_keepalive_options=None,
                  socket_type=0, retry_on_timeout=False,
                  encoding='utf-8', encoding_errors='strict',
                  decode_responses=False, parser_class=_DummyParser,
-                 socket_read_size=65536, health_check_interval=0):
+                 socket_read_size=65536, health_check_interval=0,
+                 client_name=None):
         self.pid = os.getpid()
         self.db = db
+        self.username = username
+        self.client_name = client_name
         self.password = password
         # Allow socket attributes to be passed in and saved even if they aren't used
         self.socket_timeout = socket_timeout
@@ -2524,7 +2562,7 @@ class FakeConnection(redis.Connection):
         self.next_health_check = 0
 
     def connect(self):
-        super(FakeConnection, self).connect()
+        super().connect()
         # The selector is set in redis.Connection.connect() after _connect() is called
         self._selector = FakeSelector(self._sock)
 
@@ -2565,8 +2603,17 @@ class FakeConnection(redis.Connection):
             raise response
         return self._decode(response)
 
+    def repr_pieces(self):
+        pieces = [
+            ('server', self._server),
+            ('db', self.db)
+        ]
+        if self.client_name:
+            pieces.append(('client_name', self.client_name))
+        return pieces
 
-class FakeRedisMixin(object):
+
+class FakeRedisMixin:
     def __init__(self, host='localhost', port=6379,
                  db=0, password=None, socket_timeout=None,
                  socket_connect_timeout=None,
@@ -2604,20 +2651,28 @@ class FakeRedisMixin(object):
                 'server': server
             }
             connection_pool = redis.connection.ConnectionPool(**kwargs)
-        super(FakeRedisMixin, self).__init__(
-            host, port, db, password, socket_timeout, socket_connect_timeout,
-            socket_keepalive, socket_keepalive_options, connection_pool,
-            unix_socket_path, encoding, encoding_errors, charset, errors,
-            decode_responses, retry_on_timeout,
-            ssl, ssl_keyfile, ssl_certfile, ssl_cert_reqs, ssl_ca_certs,
-            max_connections)
+        # These need to be passed by name due to
+        # https://github.com/andymccurdy/redis-py/issues/1276
+        super().__init__(
+            host=host, port=port, db=db, password=password, socket_timeout=socket_timeout,
+            socket_connect_timeout=socket_connect_timeout,
+            socket_keepalive=socket_keepalive,
+            socket_keepalive_options=socket_keepalive_options,
+            connection_pool=connection_pool,
+            unix_socket_path=unix_socket_path,
+            encoding=encoding, encoding_errors=encoding_errors,
+            charset=charset, errors=errors,
+            decode_responses=decode_responses, retry_on_timeout=retry_on_timeout,
+            ssl=ssl, ssl_keyfile=ssl_keyfile, ssl_certfile=ssl_certfile,
+            ssl_cert_reqs=ssl_cert_reqs, ssl_ca_certs=ssl_ca_certs,
+            max_connections=max_connections)
 
     @classmethod
     def from_url(cls, url, db=None, **kwargs):
         server = kwargs.pop('server', None)
         if server is None:
             server = FakeServer()
-        self = super(FakeRedisMixin, cls).from_url(url, db, **kwargs)
+        self = super().from_url(url, db, **kwargs)
         # Now override how it creates connections
         pool = self.connection_pool
         pool.connection_class = FakeConnection
